/*
 * dvbstub1_input.c
 *
 * Copyright (C) 2009 Sergey Mironov <ierton@gmail.com>
 *
 * based on pluto2.c 1.10 - http://instinct-wp8.no-ip.org/pluto/
 * 	by Dany Salman <salmandany@yahoo.fr>
 *	Copyright (c) 2004 TDF
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/cdev.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/fs.h>		/* everything... */
#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* copy_*_user */

#include "dvbstub1.h"


#ifndef DVBSTUB_OLD_KERNEL
#include <linux/semaphore.h>
#endif


struct dvbstub1_input {
	/* mutual exclusion semaphore */
	struct semaphore sem;
	/* Char device structure */
	struct cdev cdev;

	int bcalc_users;
	struct baudcalc bcalc;

	wait_queue_head_t queue;
};

static struct dvbstub1_input* g_dvbstub1_input;
static struct dvbstub1_input_ops *g_dvbstub1_input_ops;
static int g_dvbstub1_input_count = 0;

/** 
 * @brief 0 - Dynamic device number allocation
 */
static int dvbstub1_input_major = 0;
static int dvbstub1_input_minor = 0;


struct sizeable_buf { char* buf; int size; };
#define SIZEABLE_BUF_INITIALIZER {NULL,0}
struct sizeable_buf g_read_buf = SIZEABLE_BUF_INITIALIZER;
struct sizeable_buf g_write_buf = SIZEABLE_BUF_INITIALIZER;
int szbuf_checked_alloc(struct sizeable_buf* buf, size_t sz)
{
	if(!buf->buf || (sz > buf->size)) {
		if(buf->buf)
			kfree(buf->buf);
		buf->buf = kmalloc(sz, GFP_KERNEL);
		if(!buf->buf) {
			buf->size = 0;
			return -ENOMEM;
		}
		buf->size = sz;
	}
	return 0;
}

void szbuf_checked_free(struct sizeable_buf* buf)
{
	if(buf->buf)
		kfree(buf->buf);
	buf->buf = NULL;
	buf->size = 0;
}

/** 
 * @brief Opens dvbstub input device (only writing is supported)
 */
int dvbstub1_input_open(struct inode *inode, struct file *filp)
{
	/* device information */
	struct dvbstub1_input *input = 
		container_of(inode->i_cdev, struct dvbstub1_input, cdev);

	filp->private_data = input; /* for other methods */

	/* now trim to 0 the length of the device if open was write-only */
	if ( (filp->f_flags & O_ACCMODE) != O_WRONLY) {
		ERR("Only O_WRONLY mode supported\n");
		return -EPERM;
	}

	++input->bcalc_users;

	//TODO: Allow only one simultanious open descriptor
	/* Success */
	DBG("Opening input device %d\n", input-g_dvbstub1_input);
	return 0;
}

int dvbstub1_input_release(struct inode *inode, struct file *filp)
{
	struct dvbstub1_input *input = 
		container_of(inode->i_cdev, struct dvbstub1_input, cdev);

	if(--input->bcalc_users == 0) { 
		DBG("Resetting baudcalc on device %d\n", input-g_dvbstub1_input);
		baudcalc_reset(&input->bcalc);
	}

	DBG("Releasing input device %d\n", input-g_dvbstub1_input);
	return 0;
}

/*
 * Data management: read and write
 */
ssize_t dvbstub1_input_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	return -EPERM;
}

ssize_t dvbstub1_input_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
	unsigned long free = 0;
	unsigned long delay = 0;
	unsigned long now;

	int ret = 0;
	struct dvbstub1_input *input = filp->private_data;

	ret = szbuf_checked_alloc(&g_write_buf, count);
	if(ret) {
		ERR("szbuf_checked_alloc() failes. Error code: %d.", ret);
		goto out;
	}

	ret = copy_from_user(g_write_buf.buf, buf, count);
	if (ret) {
		ERR("copy_from_user() failes. Error code: %d. Returning EFAULT", ret);
		ret = -EFAULT;
		goto out;
	}


	/* FIXME: What if device being closed while sleeping in wait_event_xxx() ? */
	while(1) {
		now = jiffies;
		free = baudcalc_test(&input->bcalc, now, count, &delay);

		if(free == 0) {
/*            INFO("Going to sleep for %lu jiffies", delay);*/
			if (wait_event_interruptible_timeout(input->queue, 0, delay)) {
				ERR("wait_event_interruptible_timeout() failes");
				return -ERESTARTSYS;
			}
		}
		else
			break;
	}

	if(g_dvbstub1_input_ops->write) {
		ret = g_dvbstub1_input_ops->write(input-g_dvbstub1_input, g_write_buf.buf, free);
		if (ret < 0) {
			ERR("g_dvbstub1_input_ops->write() failes. Error code: %d.", ret);
			goto out;
		}
	}

	baudcalc_add(&input->bcalc, ret);

	*f_pos += ret;
/*    ret = count;*/
out:
	return ret;
}

/*
 * The ioctl() implementation
 */
#if 0
int dvbstub1_input_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret;
    
	/*
	 * extract the type and number bitfields, and don't decode
	 * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok()
	 */
	if (_IOC_TYPE(cmd) != SCULL_IOC_MAGIC) return -ENOTTY;
	if (_IOC_NR(cmd) > SCULL_IOC_MAXNR) return -ENOTTY;

	switch(cmd) {

	  case SCULL_P_IOCTSIZE:
		scull_p_buffer = arg;
		break;

	  case SCULL_P_IOCQSIZE:
		return scull_p_buffer;


	  default:  /* redundant, as cmd was checked against MAXNR */
		return -ENOTTY;
	}
	return retval;

}
#endif

struct file_operations g_dvbstub1_input_fops = {
	.owner =    THIS_MODULE,
	.read =     dvbstub1_input_read,
	.write =    dvbstub1_input_write,
/*    .ioctl =    dvbstub1_input_ioctl,*/
	.open =     dvbstub1_input_open,
	.release =  dvbstub1_input_release,
};

int __init dvbstub1_input_init(int dcount, struct dvbstub1_input_ops *pops)
{
	int ret = 0;
	int i,j;
	struct dvbstub1_input* input;
	dev_t dev_region;
	dev_t dev;

	DBG("entering");

	if(dcount < 1 || dcount > 10) {
		ERR("Invalid argument count");
		return -EINVAL;
	}

	input = kmalloc(dcount*sizeof(struct dvbstub1_input), GFP_KERNEL);
	if(!input) {
		ERR("kmalloc() failed. ret = %d", ret);
		return -ENOMEM;
	}

	ret = alloc_chrdev_region(&dev_region, dvbstub1_input_minor, dcount, pops->name);
	if(ret) {
		ERR("alloc_chrdev_region() fails. error code: %d", ret);
		goto kfree_err;
	}

	dvbstub1_input_major = MAJOR(dev_region);

	for(i=0; i<dcount; ++i) {
		INFO("Registering char device %s%d", pops->name, dvbstub1_input_minor + i);

		sema_init(&input[i].sem, 1);
		cdev_init(&input[i].cdev, &g_dvbstub1_input_fops);
		baudcalc_init(&input[i].bcalc, pops->baud_rate);
		input[i].bcalc_users = 0;

		init_waitqueue_head(&input[i].queue);

		dev = MKDEV(dvbstub1_input_major, dvbstub1_input_minor + i);
		input[i].cdev.owner = THIS_MODULE;
		ret = cdev_add(&input[i].cdev, dev, 1);
		if (ret) {
			ERR("Error %d adding dvbstub1_input devive %d", ret, i);
			goto cdev_add_err;
		}
	}
	g_dvbstub1_input = input;
	g_dvbstub1_input_count = dcount;
	g_dvbstub1_input_ops = pops;
out:
	DBG("leaving");
	return ret;

cdev_add_err:
	for(j=0; j<i; ++j)
		cdev_del(&input[j].cdev);
/*unreg_chrdev_reg_err:*/
	unregister_chrdev_region(dev_region, dcount);
kfree_err:
	kfree(input);
	goto out;
}

void __exit dvbstub1_input_uninit(void)
{
	int i;
	dev_t dev_region;

	DBG("entering");
	for(i=0; i<g_dvbstub1_input_count; ++i) {
		cdev_del(&g_dvbstub1_input[i].cdev);
	}

	dev_region = MKDEV(dvbstub1_input_major, dvbstub1_input_minor);
	unregister_chrdev_region(dev_region, g_dvbstub1_input_count);

	kfree(g_dvbstub1_input);

	g_dvbstub1_input = NULL;
	g_dvbstub1_input_ops = NULL;
	g_dvbstub1_input_count = 0;

	szbuf_checked_free(&g_write_buf);
	szbuf_checked_free(&g_read_buf);
	DBG("leaving");
}


